/*
 * libfx - a library for FXS and FXO signaling
 *
 * fx.c
 *
 * Written by Steve Underwood <steveu@coppice.org>
 *
 * Copyright (C) 2003 Steve Underwood
 *
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: fx.c,v 1.49 2005/12/20 15:57:08 steveu Exp $
 */

/*! \file libfx.h */

#define _ISOC9X_SOURCE  1
#define _ISOC99_SOURCE  1

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <limits.h>
#include <stdio.h>
#include <stdarg.h>
#include <math.h>
#include <fcntl.h>
#include <stdlib.h>
#include <inttypes.h>
#include <unistd.h>
#include <string.h>
#include <strings.h>
#include <errno.h>
#include <signal.h>
#include <time.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/ioctl.h>

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <libxml/xinclude.h>

#include <zaptel/zaptel.h>

#include <tiffio.h>
#include <spandsp.h>

#include <unicall.h>
#include <unicall/hashtable.h>
#include <unicall/unicall-private.h>
#include <libsupertone.h>

#include "libfx.h"
#include "fx.h"

#define BLOCKING_RELEASE_TIME       450

static uc_check_event_t check_event;
static uc_new_t create_new;
static uc_call_control_t call_control;
static uc_error_message_t error_message;
static uc_get_device_handles_t get_device_handles;
static uc_channel_open_t channel_open;
static uc_channel_close_t channel_close;
static uc_channel_set_api_codec_t channel_set_api_codec;
static uc_channel_write_t channel_write;
static uc_channel_flush_t channel_flush;
static uc_channel_gains_t channel_gains;
static uc_channel_echo_cancel_t channel_echo_cancel;
static uc_channel_switching_t channel_switching;

uc_protocol_t protocol_descriptor =
{
    "fx",
    "FXO/FXS signalling",
    "FXO/S",
    "ls\0"
    "gs\0"
    "ks\0",
    "Loop start\0"
    "Ground start\0"
    "Kewl start\0",

    check_event,
    create_new,
    call_control,
    error_message,
    get_device_handles,
    channel_open,
    channel_close,
    channel_set_api_codec,
    channel_write,
    channel_flush,
    channel_gains,
    channel_echo_cancel,
    channel_switching
};

enum
{
    ACTIVE_NONE,
    ACTIVE_SPEECH,
    ACTIVE_SPEECH_AND_SUPER_TONE,
    ACTIVE_DTMF,
    ACTIVE_ADSI,
    ACTIVE_SUPER_TONE
};

static struct zt_ring_cadence cadences[] =
{
    {{125, 125, 2000, 4000}},                       /* Quick chirp followed by normal ring */
    {{250, 250, 500, 1000, 250, 250, 500, 4000}},   /* British style ring */
    {{125, 125, 125, 125, 125, 4000}},              /* Three short bursts */
    {{1000, 500, 2500, 5000}},                      /* Long ring */
};

#if 0
static int cid_rings[] =
{
    2,      /* Right after first long ring */
    0,      /* Before the first ring */
    5,      /* After the third chirp */
    2,      /* In the long silence */
};
#endif

#define NUM_CADENCE (sizeof(cadences)/sizeof(cadences[0]))

static void start_proceeding(uc_t *uc, uc_call_t *call);
static void seize_wait_expired(uc_t *uc, void *data);
static void local_unblocking_expired(uc_t *uc, void *data);
static void start_far_disconnected(uc_t *uc, uc_call_t *call);
static void super_tone_detected(void *user_data, int tone_id);
static void adsi_msg_pickup(void *user_data, const uint8_t *msg, int len);

#define EC_OFF      0

static int hook_control(uc_t *uc, int ch, int on_hook);
static int ring_control(uc_t *uc, int ch, int ring_on);
static int control_ring_cadence(uc_t *uc, int ch, int *cadence);
static int control_echo_cancelling(uc_t *uc, int ch, int mode);
static int pulse_dial(uc_t *uc, int ch, const char *digits, int replace);
static int check_alarms(uc_t *uc, int ch, int *alarms);

static int variant_element(const char **variant, int which)
{
    int i;
    const char *s;
    const char *t;

    /* Find element "which" in a comma separated string of elements */
    t = *variant - 1;
    for (i = -1;  i < which;  i++)
    {
        s = t + 1;
        if ((t = strchr(t + 1, ',')) == NULL)
        {
            t = s + strlen(s);
            break;
        }
        /*endif*/
    }
    /*endfor*/
    if (i >= which - 1)
    {
        *variant = s;
        return  t - s;
    }
    /*endif*/
    return 0;
}
/*- End of function --------------------------------------------------------*/

static int protocol_variant(const char *variant)
{
    int i;
    int inlen;
    int len;
    char const *s;
    char const *t;

    inlen = variant_element(&variant, 0);
    s = protocol_descriptor.variants;
    for (i = 1;  *s;  i++)
    {
        len = strlen(s);
        if (len == inlen  &&  strncmp(variant, s, len) == 0)
            return  i;
        /*endif*/
        s += (len + 1);
    }
    /*endfor*/
    return  -1;
}
/*- End of function --------------------------------------------------------*/

static void dialed_digit_pickup(void *user_data, const char *digits, int len)
{
    uc_t *uc;
    fx_signaling_state_t *fx;
    uc_call_t *call;
    uc_event_t ev;
    int ch;
    int x;
    int old_len;

    uc = (uc_t *) user_data;
    fx = (fx_signaling_state_t *) uc->signaling;
    call = fx->call;
    ch = call->chan;

    old_len = strlen(call->callparms.destination_number);
    if (old_len + len > UC_MAXPHONENUMLEN)
        len = UC_MAXPHONENUMLEN - old_len;
    /*endif*/
    memcpy(&call->callparms.destination_number[old_len], digits, len);
    call->callparms.destination_number[old_len + len] = '\0';
    if (uc->chan[ch].active_tx != ACTIVE_NONE)
    {
        uc->chan[ch].active_tx = ACTIVE_NONE;
        /* Now we've turned off the source of interference, we really don't need an
           echo canceller on this channel. */
        control_echo_cancelling(uc, ch, EC_OFF);
    }
    /*endif*/
    uc_log(uc, UC_LOG_FLOW, "DNIS now '%s'\n", call->callparms.destination_number);

    if (call->state == UC_STATE_DETECTED
        ||
        call->state == UC_STATE_OFFERED)
    {
        memset (&ev.offered, '\0', sizeof(ev.offered));
        ev.e = UC_EVENT_MOREDIGITS;
        ev.offered.channel = call->chan;
        strcpy(ev.offered.parms.originating_number, call->callparms.originating_number);
        strcpy(ev.offered.parms.destination_number, call->callparms.destination_number);
        ev.offered.parms.calling_party_category = UC_CALLER_CATEGORY_UNKNOWN;
        ev.offered.parms.call_type = 0;
        ev.offered.parms.userinfo_layer1_protocol = call->callparms.userinfo_layer1_protocol;
        ev.offered.crn = call->crn;
        ev.offered.call = call;
        uc_report_event(uc, &ev, sizeof(ev));
    }
    /*endif*/
}
/*- End of function --------------------------------------------------------*/

static void adsi_msg_pickup(void *user_data, const uint8_t *msg, int len)
{
    int i;
    int l;
    uint8_t field_type;
    const uint8_t *field_body;
    int field_len;
    uint8_t body[256];
    char log_msg[256];
    char *s;
    int ch;
    int message_type;
    uc_t *uc;
    fx_signaling_state_t *fx;
    uc_call_t *call;

    uc = (uc_t *) user_data;
    fx = (fx_signaling_state_t *) uc->signaling;
    call = fx->call;
    ch = call->chan;
    
    uc_log(uc, UC_LOG_FLOW, "ADSI message received (%d bytes)\n", len);

    s = log_msg;
    for (i = 0;  i < len;  i++)
    {
        sprintf(s, "%02x ", msg[i]);
        s += 3;
        if ((i & 0xf) == 0xf)
        {
            uc_log(uc, UC_LOG_FLOW, log_msg);
            s = log_msg;
        }
        /*endif*/
    }
    /*endfor*/
    strcpy(s, "\n");
    uc_log(uc, UC_LOG_FLOW, log_msg);
    l = -1;
    do
    {
        l = adsi_next_field(&fx->adsi_rx_state, msg, len, l, &field_type, &field_body, &field_len);
        if (l > 0)
        {
            if (field_body)
            {
                memcpy(body, field_body, field_len);
                body[field_len] = '\0';
                uc_log(uc, UC_LOG_FLOW, "Type %x, len %d, '%s'\n", field_type, field_len, body);
                switch (fx->adsi_tx_state.standard)
                {
                case ADSI_STANDARD_CLASS:
                    if (message_type == CLASS_MDMF_CALLERID)
                    {
                        switch (field_type)
                        {
                        case MCLASS_DATETIME:
                            break;
                        case MCLASS_CALLER_NUMBER:
                            memcpy(call->callparms.originating_number, field_body, field_len);
                            call->callparms.originating_number[field_len] = '\0';
                            break;
                        case MCLASS_CALLER_NAME:
                            break;
                        case MCLASS_DIALED_NUMBER:
                            memcpy(call->callparms.destination_number, field_body, field_len);
                            call->callparms.destination_number[field_len] = '\0';
                            break;
                        }
                        /*endswitch*/
                    }
                    /*endif*/
                    break;
                case ADSI_STANDARD_CLIP:
                    if (message_type == CLIP_MDMF_CALLERID)
                    {
                        switch (field_type)
                        {
                        case CLIP_DATETIME:
                            break;
                        case CLIP_CALLER_NUMBER:
                            memcpy(call->callparms.originating_number, field_body, field_len);
                            call->callparms.originating_number[field_len] = '\0';
                            break;
                        case CLIP_CALLER_NAME:
                            break;
                        case CLIP_DIALED_NUMBER:
                            memcpy(call->callparms.destination_number, field_body, field_len);
                            call->callparms.destination_number[field_len] = '\0';
                            break;
                        }
                        /*endswitch*/
                    }
                    /*endif*/
                    break;
                case ADSI_STANDARD_ACLIP:
                    if (message_type == ACLIP_MDMF_CALLERID)
                    {
                        switch (field_type)
                        {
                        case ACLIP_DATETIME:
                            break;
                        case ACLIP_CALLER_NUMBER:
                            memcpy(call->callparms.originating_number, field_body, field_len);
                            call->callparms.originating_number[field_len] = '\0';
                            break;
                        case ACLIP_CALLER_NAME:
                            break;
                        case ACLIP_DIALED_NUMBER:
                            memcpy(call->callparms.destination_number, field_body, field_len);
                            call->callparms.destination_number[field_len] = '\0';
                            break;
                        }
                        /*endswitch*/
                    }
                    /*endif*/
                    break;
                case ADSI_STANDARD_JCLIP:
                    if (message_type == JCLIP_MDMF_CALLERID)
                    {
                        switch (field_type)
                        {
                        case JCLIP_CALLER_NUMBER:
                            memcpy(call->callparms.originating_number, field_body, field_len);
                            call->callparms.originating_number[field_len] = '\0';
                            break;
                        case JCLIP_CALLER_NUM_DES:
                            if (field_len == 3)
                            {
                                call->callparms.originating_ton = field_body[0] - 0x30;
                                call->callparms.originating_npi = (field_body[1] - 0x30)*10 + (field_body[2] - 0x30);
                            }
                            /*endif*/
                            break;
                        case JCLIP_DIALED_NUMBER:
                            memcpy(call->callparms.destination_number, field_body, field_len);
                            call->callparms.destination_number[field_len] = '\0';
                            break;
                        case JCLIP_DIALED_NUM_DES:
                            if (field_len == 3)
                            {
                                call->callparms.destination_ton = field_body[0] - 0x30;
                                call->callparms.destination_npi = (field_body[1] - 0x30)*10 + (field_body[2] - 0x30);
                            }
                            /*endif*/
                            break;
                        }
                        /*endswitch*/
                    }
                    /*endif*/
                    break;
                case ADSI_STANDARD_CLIP_DTMF:
                    switch (field_type)
                    {
                    case CLIP_DTMF_CALLER_NUMBER:
                        memcpy(call->callparms.originating_number, field_body, field_len);
                        call->callparms.originating_number[field_len] = '\0';
                        break;
                    case CLIP_DTMF_ABSENCE1:
                        /* This will be a reason why there is no originating number */
                        break;
                    }
                    break;
                }
                /*endswitch*/
            }
            else
            {
                message_type = field_type;
                uc_log(uc, UC_LOG_FLOW, "Message type %x\n", field_type);
                switch (fx->adsi_tx_state.standard)
                {
                case ADSI_STANDARD_CLASS:
                    if (field_type == CLASS_MDMF_CALLERID)
                        uc_log(uc, UC_LOG_FLOW, "Message type CLI\n");
                    /*endif*/
                    break;
                case ADSI_STANDARD_CLIP:
                    if (field_type == CLIP_MDMF_CALLERID)
                        uc_log(uc, UC_LOG_FLOW, "Message type CLI\n");
                    /*endif*/
                    break;
                case ADSI_STANDARD_ACLIP:
                    if (field_type == ACLIP_MDMF_CALLERID)
                        uc_log(uc, UC_LOG_FLOW, "Message type CLI\n");
                    /*endif*/
                    break;
                case ADSI_STANDARD_JCLIP:
                    if (field_type == JCLIP_MDMF_CALLERID)
                        uc_log(uc, UC_LOG_FLOW, "Message type CLI\n");
                    /*endif*/
                    break;
                case ADSI_STANDARD_CLIP_DTMF:
                    /* There is nothing to do here - DTMF messages are very simple! */
                    break;
                }
                /*endswitch*/
            }
            /*endif*/
        }
        /*endif*/
    }
    while (l > 0);
    if (fx->adsi_tx_state.standard == ADSI_STANDARD_JCLIP)
        hook_control(uc, ch, TRUE);
    /*endif*/
    uc_log(uc, UC_LOG_FLOW, "\n");
    start_proceeding(uc, fx->call);
}
/*- End of function --------------------------------------------------------*/

static void start_detected(uc_t *uc, int ch)
{
    fx_signaling_state_t *fx;
    uc_event_t ev;
    uc_call_t *call;

    fx = (fx_signaling_state_t *) uc->signaling;
    uc_log(uc, UC_LOG_FLOW, "Detected\n");

    if ((call = uc_createcall(uc, 0)) == NULL)
        return;
    /*endif*/
    fx->call = call;

    call->incoming_call = TRUE;

    call->callparms.originating_number[0] =
    call->callparms.destination_number[0] = '\0';
    fx->dropped = FALSE;

    /* Set a default for the final disconnect */
    call->disconnect_reason = UC_CAUSE_UNSPECIFIED_CAUSE;

    call->chan = ch;
    if (call->chan >= 0)
        uc->chan[call->chan].state = UC_STATE_DETECTED;
    /*endif*/
    if (uc->protocol_mode == UC_MODE_CPE)
    {
        fx->super_tone_rx_state = super_tone_rx_init(fx->super_tone_rx_state, fx->super_tones->tone_detector, super_tone_detected, uc);
        uc->chan[ch].active_rx = ACTIVE_SUPER_TONE;
        if (fx->adsi_tx_state.standard != ADSI_STANDARD_NONE  &&  uc->chan[ch].active_rx != ACTIVE_ADSI)
        {
uc_log(uc, UC_LOG_FLOW, "Starting ADSI rx between rings\n");
            /* A receiver for the ADSI/CLI standards which send digits between rings can be
               started now */
            adsi_rx_init(&fx->adsi_rx_state, fx->adsi_tx_state.standard, adsi_msg_pickup, uc);
            uc->chan[ch].active_rx = ACTIVE_ADSI;
        }
        /*endif*/
    }
    else
    {
        /* We might now expect some DTMF or pulse dialed digits */
        dtmf_rx_init(&fx->dtmf_rx_state, dialed_digit_pickup, uc);
        uc->chan[ch].active_rx = ACTIVE_DTMF;
    }
    /*endif*/
    call->state = UC_STATE_DETECTED;
    /* We need to choose the codec used by the card. */
    call->callparms.userinfo_layer1_protocol = UC_LAYER_1_ULAW;

    ev.e = UC_EVENT_DETECTED;
    ev.gen.channel = call->chan;
    ev.gen.crn = call->crn;
    ev.gen.call = call;
    uc_report_event(uc, &ev, sizeof(uc_event_generic_t));
}
/*- End of function --------------------------------------------------------*/

static void start_call_disconnected(uc_t *uc, uc_call_t *call)
{
    fx_signaling_state_t *fx;
    uc_event_t ev;
    int ch;
    
    ch = 0;
    fx = (fx_signaling_state_t *) uc->signaling;
    uc_log(uc,
           UC_LOG_FLOW,
           "Call disconnected(cause=%s [%d]) - state 0x%X\n",
           uc_cause2str(call->disconnect_reason),
           call->disconnect_reason,
           call->state);

    uc->chan[ch].active_rx =
    uc->chan[ch].active_tx = ACTIVE_NONE;

    if (call->chan >= 0)
        uc->chan[call->chan].state = UC_STATE_CALL_DISCONNECTED;
    /*endif*/
    call->state = UC_STATE_CALL_DISCONNECTED;

    ev.fardisconnected.e = UC_EVENT_DROPCALL;
    ev.fardisconnected.channel = call->chan;
    ev.fardisconnected.cause = call->disconnect_reason;
    ev.fardisconnected.crn = call->crn;
    ev.fardisconnected.call = call;
    uc_report_event(uc, &ev, sizeof(uc_event_fardisconnected_t));
}
/*- End of function --------------------------------------------------------*/

static void start_far_disconnected(uc_t *uc, uc_call_t *call)
{
    uc_event_t ev;

    uc_log(uc,
           UC_LOG_FLOW,
           "Far end disconnected(cause=%s [%d]) - state 0x%X\n",
           uc_cause2str(call->disconnect_reason),
           call->disconnect_reason,
           call->state);
    if (call->chan >= 0)
        uc->chan[call->chan].state = UC_STATE_FAR_DISCONNECTED;
    /*endif*/
    call->state = UC_STATE_FAR_DISCONNECTED;

    ev.fardisconnected.e = UC_EVENT_FARDISCONNECTED;
    ev.fardisconnected.channel = call->chan;
    ev.fardisconnected.cause = call->disconnect_reason;
    ev.fardisconnected.crn = call->crn;
    ev.fardisconnected.call = call;
    uc_report_event(uc, &ev, sizeof(uc_event_fardisconnected_t));
}
/*- End of function --------------------------------------------------------*/

static void start_idle(uc_t *uc, uc_call_t *call)
{
    fx_signaling_state_t *fx;
    uc_event_t ev;

    fx = (fx_signaling_state_t *) uc->signaling;
    /* N.B. This is really only for starting the idle state from a call. From
       initialisation, unblocking, or error recovery this is not the right
       choice. */
    uc->chan[call->chan].active_rx =
    uc->chan[call->chan].active_tx = ACTIVE_NONE;

    if (call->chan >= 0)
        uc->chan[call->chan].state = UC_STATE_IDLE;
    /*endif*/
    call->state = UC_STATE_IDLE;

    ev.e = UC_EVENT_RELEASECALL;
    ev.gen.channel = call->chan;
    ev.gen.crn = call->crn;
    ev.gen.call = call;
    uc_report_event(uc, &ev, sizeof(uc_event_generic_t));

    if (uc->protocol_mode == UC_MODE_CPE)
    {
        /* TODO: This won't happen the first time after startup */
        if (fx->adsi_tx_state.standard > 0) //  &&  ????)
        {
            /* We need a continuously operating ADSI detector in the idle state, so
               we had better start one now. */
uc_log(uc, UC_LOG_FLOW, "Starting ADSI rx while idle\n");
            adsi_rx_init(&fx->adsi_rx_state, fx->adsi_tx_state.standard, adsi_msg_pickup, uc);
            uc->chan[call->chan].active_rx = ACTIVE_ADSI;
        }
        /*endif*/
    }
    /*endif*/

    /* Although we destroy the call here, the event sent just above may not
       have been delivered. It is important, therefore, that the application
       only treat the call as a reference tag, and does not try to look at
       its contents. */
    uc_destroycall(uc, call->crn);
    fx->call = NULL;
}
/*- End of function --------------------------------------------------------*/

static void start_offered(uc_t *uc, uc_call_t *call)
{
    fx_signaling_state_t *fx;
    uc_event_t ev;

    fx = (fx_signaling_state_t *) uc->signaling;
    if (call->chan >= 0)
        uc->chan[call->chan].state = UC_STATE_OFFERED;
    /*endif*/
    call->state = UC_STATE_OFFERED;
    call->disconnect_reason = UC_CAUSE_NORMAL_CLEARING;

    memset (&ev.offered, '\0', sizeof(ev.offered));
    ev.offered.e = UC_EVENT_OFFERED;
    ev.offered.channel = call->chan;
    strcpy (ev.offered.parms.originating_number, call->callparms.originating_number);
    strcpy (ev.offered.parms.destination_number, call->callparms.destination_number);
    ev.offered.parms.calling_party_category = UC_CALLER_CATEGORY_UNKNOWN;
    ev.offered.parms.call_type = 0;
    ev.offered.parms.userinfo_layer1_protocol = call->callparms.userinfo_layer1_protocol;
    ev.offered.crn = call->crn;
    ev.offered.call = call;
    uc_report_event(uc, &ev, sizeof(uc_event_offered_t));
}
/*- End of function --------------------------------------------------------*/

static void start_proceeding(uc_t *uc, uc_call_t *call)
{
    uc_event_t ev;

    if (call->chan >= 0)
        uc->chan[call->chan].state = UC_STATE_PROCEEDING;
    /*endif*/
    call->state = UC_STATE_PROCEEDING;

    ev.e = UC_EVENT_PROCEEDING;
    ev.gen.channel = call->chan;
    ev.gen.crn = call->crn;
    ev.gen.call = call;
    uc_report_event(uc, &ev, sizeof(uc_event_generic_t));
}
/*- End of function --------------------------------------------------------*/

static void start_dialing(uc_t *uc, uc_call_t *call)
{
    uc_event_t ev;

    if (call->chan >= 0)
        uc->chan[call->chan].state = UC_STATE_DIALING;
    /*endif*/
    call->state = UC_STATE_DIALING;

    ev.e = UC_EVENT_DIALING;
    ev.gen.channel = call->chan;
    ev.gen.crn = call->crn;
    ev.gen.call = call;
    uc_report_event(uc, &ev, sizeof(uc_event_generic_t));
}
/*- End of function --------------------------------------------------------*/

static void start_alerting(uc_t *uc, uc_call_t *call)
{
    uc_event_t ev;

    if (call->chan >= 0)
        uc->chan[call->chan].state = UC_STATE_ALERTING;
    /*endif*/
    call->state = UC_STATE_ALERTING;

    ev.e = UC_EVENT_ALERTING;
    ev.gen.channel = call->chan;
    ev.gen.crn = call->crn;
    ev.gen.call = call;
    uc_report_event(uc, &ev, sizeof(uc_event_generic_t));
}
/*- End of function --------------------------------------------------------*/

static void start_connected(uc_t *uc, uc_call_t *call)
{
    fx_signaling_state_t *fx;
    uc_event_t ev;

    fx = (fx_signaling_state_t *) uc->signaling;

    /* Set the default for the clear that will occur later */
    call->disconnect_reason = UC_CAUSE_NORMAL_CLEARING;
    if (call->chan >= 0)
    {
        uc->chan[call->chan].state = UC_STATE_CONNECTED;
        if (uc->protocol_mode == UC_MODE_CPE)
            uc->chan[call->chan].active_rx = ACTIVE_SPEECH_AND_SUPER_TONE;
        else
            uc->chan[call->chan].active_rx = ACTIVE_SPEECH;
        /*endif*/
        uc->chan[call->chan].active_tx = ACTIVE_SPEECH;
    }
    /*endif*/
    call->state = UC_STATE_CONNECTED;

    if (call->incoming_call)
    {
        ev.e = UC_EVENT_ANSWERED;
    }
    else
    {
        if (fx->wait_for_answer_timer >= 0)
        {
            uc_schedule_del(uc, fx->wait_for_answer_timer);
            fx->wait_for_answer_timer = -1;
        }
        /*endif*/
        ev.e = UC_EVENT_CONNECTED;
    }
    /*endif*/
    ev.gen.channel = call->chan;
    ev.gen.crn = call->crn;
    ev.gen.call = call;
    uc_report_event(uc, &ev, sizeof(uc_event_generic_t));
}
/*- End of function --------------------------------------------------------*/

static void seize_wait_expired(uc_t *uc, void *data)
{
    fx_signaling_state_t *fx;
    int ch;
    int x;

    ch = (intptr_t) data;
    fx = (fx_signaling_state_t *) uc->signaling;
    /* Now the line should be stable, and we can dial. */
    uc_log(uc, UC_LOG_FLOW, "seize_wait_expired\n");
    fx->seize_wait_timer = -1;

    if (fx->call->callparms.destination_number[0])
    {
        uc_log(uc, UC_LOG_FLOW, "Dialing '%s'\n", fx->call->callparms.destination_number);
        if (fx->pulse_dial)
        {
            if (pulse_dial(uc, ch, fx->call->callparms.destination_number, TRUE) != UC_RET_OK)
            {
                /* Dialing failed. */
                hook_control(uc, ch, TRUE);
            }
            /*endif*/
        }
        else
        {
            dtmf_put(&fx->dtmf_tx_state, fx->call->callparms.destination_number);
            uc->chan[ch].active_rx =
            uc->chan[ch].active_tx = ACTIVE_DTMF;
        }
        /*endif*/
    }
    /*endif*/
}
/*- End of function --------------------------------------------------------*/

static void local_unblocking_expired(uc_t *uc, void *data)
{
    fx_signaling_state_t *fx;
    uc_event_t ev;
    int ch;
    
    ch = (int) (intptr_t) data;
    fx = (fx_signaling_state_t *) uc->signaling;
    uc_log(uc, UC_LOG_FLOW, "local_unblocking_expired\n");
    fx->local_unblocking_timer = -1;

    if (uc->chan[ch].state == UC_STATE_BLOCKED)
        uc->chan[ch].state = UC_STATE_IDLE;
    /*endif*/
    fx->local_blocked = FALSE;

    if (fx->adsi_tx_state.standard > 0) //  &&  ????)
    {
        /* We need a continuously operating ADSI detector in the idle state, so
           we had better start one now. */
uc_log(uc, UC_LOG_FLOW, "Starting ADSI rx on unblock\n");
        adsi_rx_init(&fx->adsi_rx_state, fx->adsi_tx_state.standard, adsi_msg_pickup, uc);
        uc->chan[ch].active_rx = ACTIVE_ADSI;
    }
    /*endif*/

    ev.e = UC_EVENT_LOCALUNBLOCKED;
    ev.gen.channel = ch;
    ev.gen.crn = 0;
    ev.gen.call = NULL;
    uc_report_event(uc, &ev, sizeof(uc_event_generic_t));

    /* There is no formal blocking procedure with these protocols, so pretend
       the far end has unblocked. */
    ev.e = UC_EVENT_FARUNBLOCKED;
    ev.gen.channel = ch;
    ev.gen.crn = 0;
    ev.gen.call = NULL;
    uc_report_event(uc, &ev, sizeof(uc_event_generic_t));
}
/*- End of function --------------------------------------------------------*/

static void super_tone_detected(void *user_data, int tone_id)
{
    int ch;
    uc_t *uc;
    fx_signaling_state_t *fx;
    uc_event_t ev;
    int x;

    uc = (uc_t *) user_data;
    ch = 0;
    fx = (fx_signaling_state_t *) uc->signaling;
    
    uc_log(uc, UC_LOG_FLOW, "Tone %d '%s' detected\n", tone_id, (tone_id >= 0)  ?  fx->super_tones->tone_names[tone_id]  :  "None");
    if (tone_id >= 0)
    {
        switch (fx->super_tones->tone_types[tone_id])
        {
        case ST_TYPE_DIALTONE:
            if (uc->chan[ch].state == UC_STATE_DIALING)
            {
                memset (&ev.gen, '\0', sizeof(ev.gen));
                ev.e = UC_EVENT_DIALTONE;
                ev.gen.channel = fx->call->chan;
                ev.gen.crn = fx->call->crn;
                ev.gen.call = fx->call;
                uc_report_event(uc, &ev, sizeof(uc_event_generic_t));

                if (fx->call->callparms.destination_number[0])
                {
                    uc_log(uc, UC_LOG_FLOW, "Dialing '%s'\n", fx->call->callparms.destination_number);
                    if (fx->pulse_dial)
                    {
                        if (pulse_dial(uc, ch, fx->call->callparms.destination_number, TRUE) != UC_RET_OK)
                        {
                            /* Dialing failed. */
                            hook_control(uc, ch, TRUE);
                        }
                        /*endif*/
                    }
                    else
                    {
                        dtmf_put(&fx->dtmf_tx_state, fx->call->callparms.destination_number);
                        //uc->chan[ch].active_rx = ACTIVE_NONE;
                        uc->chan[ch].active_tx = ACTIVE_DTMF;
                    }
                    /*endif*/
                }
                /*endif*/
            }
            /*endif*/
            break;
        case ST_TYPE_RINGBACK:
            start_alerting(uc, fx->call);
            break;
        case ST_TYPE_BUSY:
            if (uc->chan[ch].state == UC_STATE_CONNECTED)
                fx->call->disconnect_reason = UC_CAUSE_NORMAL_CLEARING;
            else
                fx->call->disconnect_reason = UC_CAUSE_USER_BUSY;
            /*endif*/
            start_far_disconnected(uc, fx->call);
            break;
        case ST_TYPE_CONGESTED:
            fx->call->disconnect_reason = UC_CAUSE_NETWORK_CONGESTION;
            start_far_disconnected(uc, fx->call);
            break;
        case ST_TYPE_NU:
            fx->call->disconnect_reason = UC_CAUSE_UNASSIGNED_NUMBER;
            start_far_disconnected(uc, fx->call);
            break;
        case ST_TYPE_DISCONNECTED:
            if (uc->chan[ch].state == UC_STATE_CONNECTED)
            {
                fx->call->disconnect_reason = UC_CAUSE_NORMAL_CLEARING;
                start_far_disconnected(uc, fx->call);
            }
            /*endif*/
            break;
        }
        /*endswitch*/
    }
    else
    {
        if (uc->chan[ch].state == UC_STATE_ALERTING)
            start_connected(uc, fx->call);
        /*endif*/
    }
    /*endif*/
}
/*- End of function --------------------------------------------------------*/

static int build_cli_msg(uc_t *uc, uint8_t *msg) 
{
    fx_signaling_state_t *fx;
    int ch;
    time_t now;
    struct tm *tm;
    int len;
    uint8_t ie_data[8 + 1];
    
    ch = 0;
    fx = (fx_signaling_state_t *) uc->signaling;

    len = 0;
    switch (fx->adsi_tx_state.standard)
    {
    case ADSI_STANDARD_CLASS:
        len = adsi_add_field(&fx->adsi_tx_state, msg, -1, CLASS_MDMF_CALLERID, NULL, 0);
        time(&now);
        tm = localtime(&now);
        snprintf((char *) ie_data,
                 9,
                 "%02d%02d%02d%02d",
                 tm->tm_mon + 1,
                 tm->tm_mday,
                 tm->tm_hour,
                 tm->tm_min);
        len = adsi_add_field(&fx->adsi_tx_state, msg, len, MCLASS_DATETIME, ie_data, 8);
        if (fx->call->callparms.originating_number[0])
            len = adsi_add_field(&fx->adsi_tx_state, msg, len, MCLASS_CALLER_NUMBER, (const uint8_t *) fx->call->callparms.originating_number, strlen(fx->call->callparms.originating_number));
        /*endif*/
        if (fx->call->callparms.originating_name[0])
            len = adsi_add_field(&fx->adsi_tx_state, msg, len, MCLASS_CALLER_NAME, (const uint8_t *) fx->call->callparms.originating_name, strlen(fx->call->callparms.originating_name));
        /*endif*/
        if (fx->call->callparms.destination_number[0])
            len = adsi_add_field(&fx->adsi_tx_state, msg, len, MCLASS_DIALED_NUMBER, (const uint8_t *) fx->call->callparms.destination_number, strlen(fx->call->callparms.destination_number));
        /*endif*/
        break;
    case ADSI_STANDARD_CLIP:
        len = adsi_add_field(&fx->adsi_tx_state, msg, -1, CLIP_MDMF_CALLERID, NULL, 0);
        time(&now);
        tm = localtime(&now);
        snprintf((char *) ie_data,
                 9,
                 "%02d%02d%02d%02d",
                 tm->tm_mon + 1,
                 tm->tm_mday,
                 tm->tm_hour,
                 tm->tm_min);
        len = adsi_add_field(&fx->adsi_tx_state, msg, len, CLIP_DATETIME, ie_data, 8);
        if (fx->call->callparms.originating_number[0])
            len = adsi_add_field(&fx->adsi_tx_state, msg, len, CLIP_CALLER_NUMBER, (const uint8_t *) fx->call->callparms.originating_number, strlen(fx->call->callparms.originating_number));
        /*endif*/
        if (fx->call->callparms.originating_name[0])
            len = adsi_add_field(&fx->adsi_tx_state, msg, len, CLIP_CALLER_NAME, (const uint8_t *) fx->call->callparms.originating_name, strlen(fx->call->callparms.originating_name));
        /*endif*/
        if (fx->call->callparms.destination_number[0])
            len = adsi_add_field(&fx->adsi_tx_state, msg, len, CLIP_DIALED_NUMBER, (const uint8_t *) fx->call->callparms.destination_number, strlen(fx->call->callparms.destination_number));
        /*endif*/
        break;
    case ADSI_STANDARD_ACLIP:
        len = adsi_add_field(&fx->adsi_tx_state, msg, -1, ACLIP_MDMF_CALLERID, NULL, 0);
        time(&now);
        tm = localtime(&now);
        snprintf((char *) ie_data,
                 9,
                 "%02d%02d%02d%02d",
                 tm->tm_mon + 1,
                 tm->tm_mday,
                 tm->tm_hour,
                 tm->tm_min);
        len = adsi_add_field(&fx->adsi_tx_state, msg, len, ACLIP_DATETIME, ie_data, 8);
        if (fx->call->callparms.originating_number[0])
            len = adsi_add_field(&fx->adsi_tx_state, msg, len, ACLIP_CALLER_NUMBER, (const uint8_t *) fx->call->callparms.originating_number, strlen(fx->call->callparms.originating_number));
        /*endif*/
        if (fx->call->callparms.originating_name[0])
            len = adsi_add_field(&fx->adsi_tx_state, msg, len, ACLIP_CALLER_NAME, (const uint8_t *) fx->call->callparms.originating_name, strlen(fx->call->callparms.originating_name));
        /*endif*/
        if (fx->call->callparms.destination_number[0])
            len = adsi_add_field(&fx->adsi_tx_state, msg, len, ACLIP_DIALED_NUMBER, (const uint8_t *) fx->call->callparms.destination_number, strlen(fx->call->callparms.destination_number));
        /*endif*/
        break;
    case ADSI_STANDARD_JCLIP:
        len = adsi_add_field(&fx->adsi_tx_state, msg, -1, JCLIP_MDMF_CALLERID, NULL, 0);
        if (fx->call->callparms.originating_number[0])
        {
            len = adsi_add_field(&fx->adsi_tx_state, msg, len, JCLIP_CALLER_NUMBER, (const uint8_t *) fx->call->callparms.originating_number, strlen(fx->call->callparms.originating_number));
            ie_data[0] = 0x30 + fx->call->callparms.destination_ton;
            ie_data[1] = 0x30 + fx->call->callparms.destination_npi/10;
            ie_data[2] = 0x30 + fx->call->callparms.destination_npi%10;
            len = adsi_add_field(&fx->adsi_tx_state, msg, len, JCLIP_CALLER_NUM_DES, ie_data, 3);
        }
        /*endif*/
        if (fx->call->callparms.destination_number[0])
        {
            len = adsi_add_field(&fx->adsi_tx_state, msg, len, JCLIP_DIALED_NUMBER, (const uint8_t *) fx->call->callparms.destination_number, strlen(fx->call->callparms.destination_number));
            ie_data[0] = 0x30 + fx->call->callparms.originating_ton;
            ie_data[1] = 0x30 + fx->call->callparms.originating_npi/10;
            ie_data[2] = 0x30 + fx->call->callparms.originating_npi%10;
            len = adsi_add_field(&fx->adsi_tx_state, msg, len, JCLIP_DIALED_NUM_DES, ie_data, 3);
        }
        /*endif*/
        break;
    case ADSI_STANDARD_CLIP_DTMF:
        len = adsi_add_field(&fx->adsi_tx_state, msg, 0, CLIP_DTMF_CALLER_NUMBER, (const uint8_t *) fx->call->callparms.originating_number, strlen(fx->call->callparms.originating_number));
        break;
    }
    /*endswitch*/
    return len;
}
/*- End of function --------------------------------------------------------*/

static void process_signaling_event(uc_t *uc, int x)
{
    fx_signaling_state_t *fx;
    uc_event_t ev;
    char dig;
    int ch;
    int len;
    int changes;
    uint8_t cli_msg[256];
    int alarms;

    ch = 0;
    fx = (fx_signaling_state_t *) uc->signaling;
    uc_log(uc,
           UC_LOG_DEBUG_2,
           "Rx event 0x%08X [%8x]\n",
           x,
           uc->chan[ch].state);
    switch (x & 0xFFFF)
    {
    case ZT_EVENT_ALARM:
    case ZT_EVENT_NOALARM:
        if (check_alarms(uc, ch, &alarms) != UC_RET_OK)
        {
            /* TODO: */
        }
        /*endif*/
        changes = fx->alarms ^ alarms;
        ev.alarm.e = UC_EVENT_ALARM;
        ev.alarm.channel = ch;
        ev.alarm.raised = changes & alarms;
        ev.alarm.cleared = changes & ~alarms;
        fx->alarms = alarms;
        uc_report_event(uc, &ev, sizeof(ev));
        break;
    case ZT_EVENT_ONHOOK:
        uc_log(uc, UC_LOG_FLOW, "On hook\n");
        if (uc->protocol_mode == UC_MODE_CPE)
        {
            if (uc->chan[ch].state != UC_STATE_IDLE)
                start_idle(uc, fx->call);
            /*endif*/
        }
        else
        {
            if (uc->chan[ch].state != UC_STATE_IDLE)
            {
                if (fx->dropped)
                    start_call_disconnected(uc, fx->call);
                else
                    start_far_disconnected(uc, fx->call);
                /*endif*/
            }
            /*endif*/
        }
        /*endif*/
        break;
    case ZT_EVENT_RINGOFFHOOK:
        uc_log(uc, UC_LOG_FLOW, "Ring/off hook\n");
        if (uc->protocol_mode == UC_MODE_CPE)
        {
            /* This should be a ringing event */
            switch (uc->chan[ch].state)
            {
            case UC_STATE_IDLE:
                /* We have detected an incoming call. However,
                   if we expect CLI, we haven't reached the
                   offered stage yet. */
                start_detected(uc, ch);
                if (fx->adsi_tx_state.standard != ADSI_STANDARD_NONE)
                {
                    if (fx->adsi_tx_state.standard == ADSI_STANDARD_JCLIP)
                        hook_control(uc, ch, FALSE);
                    /*endif*/
                    fx->cid_rings = 3;
                }
                else
                {
                    /* Go directly to offered if CLI is not expected, otherwise wait */
                    start_offered(uc, fx->call);
                }
                /*endif*/
                break;
            case UC_STATE_DETECTED:
            case UC_STATE_PROCEEDING:
            case UC_STATE_ACCEPTED:
                /* If too many rings have passed, and we have not received the CLI, we
                   are never going to. */
                if (--fx->cid_rings == 0)
                    start_offered(uc, fx->call);
                /*endif*/
                break;
            }            
            /*endswitch*/
        }
        else
        {
            /* This should be the far phone going off-hook */
            if (uc->chan[ch].state == UC_STATE_ALERTING)
            {
                /* Stop the ringing. */
                ring_control(uc, ch, FALSE);
                start_connected(uc, fx->call);
            }
            else
            {
                /* Start dial tone */
                super_tone_tx_init(&fx->super_tone_tx_state, fx->super_tones->tone[ST_TYPE_DIALTONE]);
                uc->chan[ch].active_tx = ACTIVE_SUPER_TONE;
                control_echo_cancelling(uc, ch, 32);
                start_detected(uc, ch);
            }
            /*endif*/
        }
        /*endif*/
        break;
    case ZT_EVENT_WINKFLASH:
        uc_log(uc, UC_LOG_FLOW, "Wink flash\n");
        break;
    case ZT_EVENT_ABORT:
        uc_log(uc, UC_LOG_FLOW, "Abort\n");
        break;
    case ZT_EVENT_OVERRUN:
        uc_log(uc, UC_LOG_FLOW, "Overrun\n");
        break;
    case ZT_EVENT_DIALCOMPLETE:
        /* This should be for pulse dialing */
        uc_log(uc, UC_LOG_FLOW, "Dial complete\n");
        start_proceeding(uc, fx->call);
        break;
    case ZT_EVENT_RINGERON:
        /* We are ringing the far end */
        uc_log(uc, UC_LOG_FLOW, "Ringer on\n");
        break;
    case ZT_EVENT_RINGEROFF:
        /* We have stopped ringing the far end */
        uc_log(uc, UC_LOG_FLOW, "Ringer off\n");
        if (fx->adsi_tx_state.standard != ADSI_STANDARD_NONE  &&  --fx->cid_rings == 0)
        {
            uc_log(uc, UC_LOG_FLOW, "Time for the CLI\n");
            adsi_tx_init(&fx->adsi_tx_state, fx->adsi_tx_state.standard);
            len = build_cli_msg(uc, cli_msg);
            adsi_put_message(&fx->adsi_tx_state, cli_msg, len);
            uc->chan[ch].active_tx = ACTIVE_ADSI;
        }
        /*endif*/
        break;
    case ZT_EVENT_HOOKCOMPLETE:
        uc_log(uc, UC_LOG_FLOW, "Hook complete\n");
        break;
    case ZT_EVENT_PULSE_START:
        uc_log(uc, UC_LOG_FLOW, "Pulse start\n");
        break;
    default:
        uc_log(uc, UC_LOG_FLOW, "Event 0x%X\n", x);
        x &= 0xFFFF;
        if (x >= 0x30  &&  x <= 0x39)
        {
            /* This should be a pulse dialed digit */
            dig = x;
            dialed_digit_pickup(uc, &dig, 1);
        }
        else
        {
            /* This is something unexpected */
            if (uc->chan[ch].channel_error)
                uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, FX_PROTOCOL_FAIL_DEVICE_IO_ERROR);
            /*endif*/
        }
        /*endif*/
        break;
    }
    /*endswitch*/
}
/*- End of function --------------------------------------------------------*/

static uc_event_t *check_event(uc_t *uc)
{
    uint8_t buf[1024];
    int16_t pcm_buf[1024];
    int16_t *pcm_ptr;
    int res;
    int len;
    int xlen;
    int maxlen;
    int i;
    int x;
//    int y;
    fx_signaling_state_t *fx;
    int ch;
    
    ch = 0;
    fx = (fx_signaling_state_t *) uc->signaling;

    for (;;)
    {
        //x = ZT_IOMUX_SIGEVENT | ZT_IOMUX_READ | ZT_IOMUX_WRITE | ZT_IOMUX_WRITEEMPTY | ZT_IOMUX_NOWAIT;
        x = ZT_IOMUX_SIGEVENT | ZT_IOMUX_READ | ZT_IOMUX_NOWAIT;
        switch (uc->chan[ch].active_tx)
        {
        case ACTIVE_SPEECH:
            if (uc->chan[ch].channel_write)
                x |= ZT_IOMUX_WRITE;
            /*endif*/
            break;
        case ACTIVE_NONE:
            break;
        default:
            x |= ZT_IOMUX_WRITE;
            break;
        }
        /*endswitch*/
        if ((res = ioctl(uc->chan[ch].handle, ZT_IOMUX, &x)))
        {
            if (uc->chan[ch].channel_error)
                uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, FX_PROTOCOL_FAIL_DEVICE_IO_ERROR);
            /*endif*/
            break;
        }
        /*endif*/
        if (x == 0)
            break;
        /*endif*/
        if ((x & ZT_IOMUX_SIGEVENT))
        {
            uc_log(uc, UC_LOG_FLOW, "Event\n");
            res = ioctl(uc->chan[ch].handle, ZT_GETEVENT, &x);
            if (res == 0  &&  x != 0)
                process_signaling_event(uc, x);
            /*endif*/
            continue;
        }
        /*endif*/
        if ((x & ZT_IOMUX_READ))
        {
            //uc_log(uc, UC_LOG_FLOW, "Read\n");
            if ((res = read(uc->chan[ch].handle, buf, sizeof(buf))) < 0)
            {
                if (errno == ELAST)
                {
                    /* This means there is some out of band stuff to deal with - alarms,
                       signaling changes, etc. */
                }
                else if (errno != EAGAIN)
                {
                    if (uc->chan[ch].channel_error)
                        uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, FX_PROTOCOL_FAIL_DEVICE_IO_ERROR);
                    /*endif*/
                }
                /*endif*/
                continue;
            }
            /*endif*/

            switch (uc->chan[ch].active_rx)
            {
            case ACTIVE_NONE:
                break;
            default:
                /* Do the tone detection */
                switch (uc->chan[ch].api_codec)
                {
                case UC_CODEC_LINEAR16:
                    len = res >> 1;
                    pcm_ptr = (int16_t *) buf;
                    break;
                case UC_CODEC_ALAW:
                    for (i = 0;  i < res;  i++)
                        pcm_buf[i] = alaw_to_linear(buf[i]);
                    /*endfor*/
                    pcm_ptr = pcm_buf;
                    len = res;
                    break;
                case UC_CODEC_ULAW:
                    for (i = 0;  i < res;  i++)
                        pcm_buf[i] = ulaw_to_linear(buf[i]);
                    /*endfor*/
                    pcm_ptr = pcm_buf;
                    len = res;
                    break;
                default:
                    uc_log(uc, UC_LOG_WARNING, "Unknown codec %d\n", uc->chan[ch].api_codec);
                    pcm_ptr = NULL;
                    break;
                }
                /*endswitch*/
                if (pcm_ptr)
                {
                    switch (uc->chan[ch].active_rx)
                    {
                    case ACTIVE_DTMF:
                        dtmf_rx(&fx->dtmf_rx_state, pcm_ptr, len);
                        break;
                    case ACTIVE_ADSI:
                        adsi_rx(&fx->adsi_rx_state, pcm_ptr, len);
                        break;
                    default:
                        super_tone_rx(fx->super_tone_rx_state, pcm_ptr, len);
                        break;
                    }
                    /*endswitch*/
                }
                /*endif*/
                if (uc->chan[ch].active_rx != ACTIVE_SPEECH_AND_SUPER_TONE)
                    break;
                /*endif*/
                /* Fall through */
            case ACTIVE_SPEECH:
                if (uc->chan[ch].channel_read)
                {
                    if (uc->chan[ch].channel_read_codec)
                        res = uc->chan[ch].channel_read_codec(uc, ch, uc->chan[ch].channel_read_codec_user_data, buf, res);
                    /*endif*/
                    uc->chan[ch].channel_read(uc, ch, uc->chan[ch].channel_read_user_data, buf, res);
                }
                /*endif*/
                break;
            }
            /*endswitch*/
            continue;
        }
        /*endif*/
        if ((x & ZT_IOMUX_WRITE))
        {
            //uc_log(uc, UC_LOG_FLOW, "Write - %d\n", uc->chan[ch].active_tx);
            switch (uc->chan[ch].active_tx)
            {
            case ACTIVE_SPEECH:
                if (uc->chan[ch].channel_write)
                {
                    maxlen = uc->chan[ch].audio_bufsize;
                    len = uc->chan[ch].channel_write(uc, ch, uc->chan[ch].channel_write_user_data, buf, maxlen);
                    if (len > 0)
                    {
                        if (uc->chan[ch].channel_write_codec)
                            len = uc->chan[ch].channel_write_codec(uc, ch, uc->chan[ch].channel_write_codec_user_data, buf, len);
                        /*endif*/
                        if ((xlen = write(uc->chan[ch].handle, buf, len)) != len)
                        {
                            if (uc->chan[ch].channel_error)
                                uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, FX_PROTOCOL_FAIL_DEVICE_IO_ERROR);
                            /*endif*/
                        }
                        /*endif*/
                    }
                    /*endif*/
                }
                /*endif*/
                break;
            case ACTIVE_NONE:
                break;
            default:
                len = -1;
                maxlen = uc->chan[ch].audio_bufsize;
                switch (uc->chan[ch].active_tx)
                {
                case ACTIVE_ADSI:
                    len = adsi_tx(&fx->adsi_tx_state, pcm_buf, maxlen);
//uc_log(uc, UC_LOG_WARNING, "------------------ADSI %d %d\n", len, uc->chan[ch].api_codec);
                    if (len <= 0)
                    {
                        uc->chan[ch].active_tx = ACTIVE_NONE;
#if 0
                        if (fx->distinctive_ring > 0  &&  cid_rings[fx->distinctive_ring - 1] == 0)
                        {
uc_log(uc, UC_LOG_WARNING, "------------------ring second\n");
                            /* This must have been ADSI before the first ring, so now we need
                               to start ringing. */
                            /* TODO: we should have a delay here, while the tx buffer flushes */
                            if (ring_control(uc, ch, TRUE) != UC_RET_OK)
                                continue;
                            /*endif*/
                        }
                        /*endif*/
#endif
                    }
                    /*endif*/
                    break;
                case ACTIVE_DTMF:
                    len = dtmf_tx(&fx->dtmf_tx_state, pcm_buf, maxlen);
                    if (len <= 0)
                    {
                        start_proceeding(uc, fx->call);
                        uc->chan[ch].active_tx = ACTIVE_NONE;
                    }
                    /*endif*/
                    break;
                default:
                    len = super_tone_tx(&fx->super_tone_tx_state, pcm_buf, maxlen);
                    break;
                }
                /*endswitch*/
                if (len > 0)
                {
                    switch (uc->chan[ch].api_codec)
                    {
                    case UC_CODEC_LINEAR16:
                        len *= 2;
                        xlen = write(uc->chan[ch].handle, pcm_buf, len);
                        break;
                    case UC_CODEC_ALAW:
                        for (i = 0;  i < len;  i++)
                            buf[i] = linear_to_alaw(pcm_buf[i]);
                        /*endfor*/
                        xlen = write(uc->chan[ch].handle, buf, len);
                        break;
                    case UC_CODEC_ULAW:
                        for (i = 0;  i < len;  i++)
                            buf[i] = linear_to_ulaw(pcm_buf[i]);
                        /*endfor*/
                        xlen = write(uc->chan[ch].handle, buf, len);
                        break;
                    default:
                        uc_log(uc, UC_LOG_WARNING, "Unknown codec %d\n", uc->chan[ch].api_codec);
                        break;
                    }
                    /*endswitch*/
                    if (xlen != len)
                    {
                        if (uc->chan[ch].channel_error)
                            uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, FX_PROTOCOL_FAIL_DEVICE_IO_ERROR);
                        /*endif*/
                    }
                    /*endif*/
                }
                /*endif*/
            }
            /*endif*/
            continue;
        }
        /*endif*/
        if ((x & ZT_IOMUX_WRITEEMPTY))
        {
            //uc_log(uc, UC_LOG_FLOW, "Write empty\n");
            continue;
        }
        /*endif*/
    }
    /*endfor*/
    return NULL;
}
/*- End of function --------------------------------------------------------*/

static int make_call(uc_t *uc, uc_makecall_t *makecall)
{
    fx_signaling_state_t *fx;
    uc_event_t ev;
    int x;
    int ch;
    uc_call_t *call;
    const uc_callparms_t *callparms;
#if 0
    int len;
    char cli_msg[256];
#endif

    ch = 0;
    fx = (fx_signaling_state_t *) uc->signaling;
    uc_log(uc, UC_LOG_FLOW, "Make call\n");

    if (!fx->outgoing_calls_allowed)
        return  UC_RET_BLOCKED;
    /*endif*/

    if (uc->chan[ch].state == UC_STATE_BLOCKED)
        return  UC_RET_BLOCKED;
    /*endif*/

    if (uc->chan[ch].state != UC_STATE_IDLE)
        return  UC_RET_BAD_STATE;
    /*endif*/

    if ((call = uc_createcall(uc, 0)) == NULL)
        return UC_RET_MEMORY_ERROR;
    /*endif*/
    makecall->crn = call->crn;
    callparms = makecall->callparms;
    
    //fx->distinctive_ring = 0;
    fx->call = call;
    strcpy(call->callparms.destination_number, callparms->destination_number);
    call->callparms.destination_ton = callparms->destination_ton;
    call->callparms.destination_npi = callparms->destination_npi;
    strcpy(call->callparms.originating_number, callparms->originating_number);
    call->callparms.originating_ton = callparms->originating_ton;
    call->callparms.originating_npi = callparms->originating_npi;
    call->chan = ch;

    if (callparms->calling_party_category <= 9)
        fx->calling_party_category = callparms->calling_party_category;
    else
        fx->calling_party_category = UC_CALLER_CATEGORY_UNKNOWN;
    /*endif*/

    if (uc->protocol_mode == UC_MODE_CPE)
    {
        if ((x = hook_control(uc, ch, FALSE)) != UC_RET_OK)
            return x;
        /*endif*/
    }
    else
    {
        /* Select the proper cadence, and ring the phone */
        x = FALSE;
        if (fx->distinctive_ring > 0  &&  fx->distinctive_ring <= NUM_CADENCE)
        {
            control_ring_cadence(uc, ch, NULL);
            fx->cid_rings = 1;
#if 0
            if (ioctl(uc->chan[ch].handle, ZT_SETCADENCE, &cadences[fx->distinctive_ring - 1]))
                uc_log(uc, UC_LOG_WARNING, "Unable to set distinctive ring cadence %d\n", fx->distinctive_ring);
            /*endif*/
            fx->cid_rings = cid_rings[fx->distinctive_ring - 1];
            if (fx->cid_rings == 0)
            {
uc_log(uc, UC_LOG_WARNING, "------------------CLI first - %d\n", fx->adsi_tx_state.standard);
                /* We should send CLI now, and delay the start of ringing until that has finished. */
                uc_log(uc, UC_LOG_FLOW, "Time for the CLI\n");
                adsi_tx_init(&fx->adsi_tx_state, fx->adsi_tx_state.standard);
                len = build_cli_msg(uc, cli_msg);
                adsi_put_message(&fx->adsi_tx_state, cli_msg, len);
                uc->chan[ch].active_tx = ACTIVE_ADSI;
fx_set_api_codec(uc, 0, 2);
            }
            else
            {
                x = TRUE;
            }
            /*endif*/
#else
            x = TRUE;
#endif
        }
        else
        {
            control_ring_cadence(uc, ch, NULL);
            fx->cid_rings = 1;
            x = TRUE;
        }
        /*endif*/
        if (x)
        {
uc_log(uc, UC_LOG_WARNING, "------------------HOOK - %d\n", x);
            if (ring_control(uc, ch, x) != UC_RET_OK)
                return  UC_RET_DEVICE_ERROR;
            /*endif*/
        }
        /*endif*/
    }
    /*endif*/
    fx->call_type = 0;

    call->disconnect_reason = UC_CAUSE_UNSPECIFIED_CAUSE;
    call->incoming_call = FALSE;
    fx->dropped = FALSE;
    if (uc->protocol_mode == UC_MODE_CPE)
    {
        fx->super_tone_rx_state = super_tone_rx_init(fx->super_tone_rx_state, fx->super_tones->tone_detector, super_tone_detected, uc);
        uc->chan[ch].active_rx = ACTIVE_SUPER_TONE;
        if (fx->call->callparms.destination_number[0])
        {
            uc_log(uc, UC_LOG_FLOW, "Will dial '%s'\n", fx->call->callparms.destination_number);
            /* Do a seize wait for the line to settle, then dial. */
            if (!fx->pulse_dial)
                dtmf_tx_init(&fx->dtmf_tx_state);
            /*endif*/
            fx->seize_wait_timer =
                uc_schedule_event(uc, fx->seize_wait, seize_wait_expired, (void *) (intptr_t) ch);
            start_dialing(uc, call);
        }
        else
        {
            uc_log(uc, UC_LOG_FLOW, "Nothing to dial\n");
            /* Nothing to dial, so the call moves immediately to proceeding */
            start_proceeding(uc, call);
        }
        /*endif*/
    }
    else
    {
        /* We always go directly from making a call to alerting, since making a call
           is just ringing the phone and waiting for answer. */
        start_alerting(uc, call);
    }
    /*endif*/
    return  UC_RET_OK;
}
/*- End of function --------------------------------------------------------*/

static int send_more_info(uc_t *uc, uc_call_t *call, uc_sendmoreinfo_t *info)
{
    return  UC_RET_UNSUPPORTED;
}
/*- End of function --------------------------------------------------------*/

static int call_ack(uc_t *uc, uc_call_t *call, uc_callack_t *callack)
{
    return  UC_RET_UNSUPPORTED;
}
/*- End of function --------------------------------------------------------*/

static int accept_call(uc_t *uc, uc_call_t *call)
{
    fx_signaling_state_t *fx;
    uc_event_t ev;
    int ch;

    ch = call->chan;
    fx = (fx_signaling_state_t *) uc->signaling;
    uc_log(uc, UC_LOG_FLOW, "Accept call\n");
    if (call == NULL)
        return  UC_RET_BAD_STATE;
    /*endif*/
    if (call->state != UC_STATE_OFFERED
        &&
        call->state != UC_STATE_DETECTED)
    {
        return  UC_RET_BAD_STATE;
    }
    /*endif*/
    if (uc->protocol_mode == UC_MODE_CPE)
    {
        /* The CPE has no way to indicate call acceptance */
    }
    else
    {
        /* Start ringback tone */
        super_tone_tx_init(&fx->super_tone_tx_state, fx->super_tones->tone[ST_TYPE_RINGBACK]);
        uc->chan[ch].active_tx = ACTIVE_SUPER_TONE;
    }
    /*endif*/

    if (call->chan >= 0)
        uc->chan[call->chan].state = UC_STATE_ACCEPTED;
    /*endif*/
    call->state = UC_STATE_ACCEPTED;

    ev.e = UC_EVENT_ACCEPTED;
    ev.gen.channel = call->chan;
    ev.gen.crn = call->crn;
    ev.gen.call = call;
    uc_report_event(uc, &ev, sizeof(uc_event_generic_t));
    return  UC_RET_OK;
}
/*- End of function --------------------------------------------------------*/

static int answer_call(uc_t *uc, uc_call_t *call)
{
    fx_signaling_state_t *fx;
    int x;
    int ch;

    ch = call->chan;
    fx = (fx_signaling_state_t *) uc->signaling;
    uc_log(uc, UC_LOG_FLOW, "Answer call\n");
    if (call == NULL)
        return  UC_RET_BAD_STATE;
    /*endif*/
    if (uc->protocol_mode == UC_MODE_CPE)
    {
        if ((x = hook_control(uc, ch, FALSE)) != UC_RET_OK)
            return  x;
        /*endif*/
    }
    else
    {
        uc->chan[ch].active_rx =
        uc->chan[ch].active_tx = ACTIVE_SPEECH;
    }
    /*endif*/
    start_connected(uc, call);
    return  UC_RET_OK;
}
/*- End of function --------------------------------------------------------*/

static int drop_call(uc_t *uc, uc_call_t *call, int cause)
{
    fx_signaling_state_t *fx;
    int x;
    int ch;
    
    ch = call->chan;
    fx = (fx_signaling_state_t *) uc->signaling;
    uc_log(uc, UC_LOG_FLOW, "Drop call(cause=%s [%d])\n", uc_cause2str(cause), cause);
    if (call == NULL)
        return  UC_RET_BAD_STATE;
    /*endif*/
    /* TODO: Check for further state conflicts */
    if (call->state == UC_STATE_NULL
        ||
        call->state == UC_STATE_IDLE
        ||
        call->state == UC_STATE_CALL_DISCONNECTED)
    {
        return  UC_RET_BAD_STATE;
    }
    /*endif*/
    if (uc->protocol_mode == UC_MODE_CPE)
    {
        if ((x = hook_control(uc, ch, TRUE)) != UC_RET_OK)
            return x;
        /*endif*/
        start_call_disconnected(uc, call);
    }
    else
    {
        switch (call->state)
        {
        case UC_STATE_FAR_DISCONNECTED:
            start_call_disconnected(uc, call);
            break;
        case UC_STATE_ALERTING:
            if ((x = ring_control(uc, ch, FALSE)) != UC_RET_OK)
                return  UC_RET_DEVICE_ERROR;
            /*endif*/
            /* TODO: wait until the ringer has really stopped */
            start_call_disconnected(uc, call);
            break;
        default:
            /* The other end is still there, so we should send a tone to them,
               and wait for them to disconnect. */
            fx->dropped = TRUE;
            uc->chan[ch].active_rx =
            uc->chan[ch].active_tx = ACTIVE_NONE;
            switch (cause)
            {
            default:
            case UC_CAUSE_USER_BUSY:
                /* Start busy tone */
                super_tone_tx_init(&fx->super_tone_tx_state, fx->super_tones->tone[ST_TYPE_BUSY]);
                uc->chan[ch].active_tx = ACTIVE_SUPER_TONE;
                break;
            case UC_CAUSE_UNASSIGNED_NUMBER:
                /* Start NU tone */
                super_tone_tx_init(&fx->super_tone_tx_state, fx->super_tones->tone[ST_TYPE_NU]);
                uc->chan[ch].active_tx = ACTIVE_SUPER_TONE;
                break;
            case UC_CAUSE_NETWORK_CONGESTION:
                /* Start congestion tone */
                super_tone_tx_init(&fx->super_tone_tx_state, fx->super_tones->tone[ST_TYPE_CONGESTED]);
                uc->chan[ch].active_tx = ACTIVE_SUPER_TONE;
                break;
            case UC_CAUSE_NORMAL_CLEARING:
                /* Start busy tone */
                super_tone_tx_init(&fx->super_tone_tx_state, fx->super_tones->tone[ST_TYPE_DISCONNECTED]);
                uc->chan[ch].active_tx = ACTIVE_SUPER_TONE;
                break;
            }
            /*endswitch*/
        }
        /*endswitch*/
    }
    /*endif*/
    return UC_RET_OK;
}
/*- End of function --------------------------------------------------------*/

static int release_call(uc_t *uc, uc_call_t *call)
{
    fx_signaling_state_t *fx;

    fx = (fx_signaling_state_t *) uc->signaling;
    uc_log(uc, UC_LOG_FLOW, "Release call\n");
    if (call == NULL)
        return  UC_RET_BAD_STATE;
    /*endif*/
    if (call->state != UC_STATE_CALL_DISCONNECTED)
        return  UC_RET_BAD_STATE;
    /*endif*/
    start_idle(uc, call);
    return  UC_RET_OK;
}
/*- End of function --------------------------------------------------------*/

static int unblock(uc_t *uc, int ch)
{
    fx_signaling_state_t *fx;
    int x;
    
    /* ch = -1 means all voice channels */
    if (ch >= uc->num_channels)
        return UC_RET_BAD_PARAMETER;
    /*endif*/
    fx = (fx_signaling_state_t *) uc->signaling;
    ch = 0;
    uc_log(uc, UC_LOG_FLOW, "Unblock\n");

    if ((x = hook_control(uc, ch, TRUE)) != UC_RET_OK)
        return x;
    /*endif*/
    /* Start a timeout to ensure we do not seize the trunk too aggressively.
       That is, before the far end has a chance to recognise the UNBLOCKED
       condition. */
    fx->local_unblocking_timer =
        uc_schedule_event(uc, BLOCKING_RELEASE_TIME, local_unblocking_expired, (void *) (intptr_t) ch);
    return  UC_RET_OK;
}
/*- End of function --------------------------------------------------------*/

static int block(uc_t *uc, int ch)
{
    fx_signaling_state_t *fx;
    uc_event_t ev;
    int x;

    /* ch = -1 means all voice channels */
    if (ch >= uc->num_channels)
        return UC_RET_BAD_PARAMETER;
    /*endif*/
    fx = (fx_signaling_state_t *) uc->signaling;
    ch = 0;
    uc_log(uc, UC_LOG_FLOW, "Block\n");

    fx->local_blocked = TRUE;
    if ((x = hook_control(uc, ch, TRUE)) != UC_RET_OK)
        return x;
    /*endif*/
    ev.e = UC_EVENT_LOCALBLOCKED;
    ev.gen.channel = ch;
    uc_report_event(uc, &ev, sizeof(uc_event_generic_t));
    return  UC_RET_OK;
}
/*- End of function --------------------------------------------------------*/

static int user_to_user(uc_t *uc, uc_call_t *call, uc_usertouser_t *msg)
{
    fx_signaling_state_t *fx;
    uc_event_t ev;
    int ch;
    
    if (msg->len <= 0)
        return UC_RET_OK;
    /*endif*/
    ch = call->chan;
    if (ch >= uc->num_channels)
        return UC_RET_BAD_PARAMETER;
    /*endif*/
    fx = (fx_signaling_state_t *) uc->signaling;
    adsi_tx_init(&fx->adsi_tx_state, fx->adsi_tx_state.standard);
    adsi_put_message(&fx->adsi_tx_state, msg->message, msg->len);
    uc->chan[ch].active_tx = ACTIVE_ADSI;
    return  UC_RET_OK;
}
/*- End of function --------------------------------------------------------*/

static int call_control(uc_t *uc, int op, uc_call_t *call, void *data)
{
    uc_log(uc, UC_LOG_FLOW, "Call control(%d)\n", op);
    switch (op)
    {
    case UC_OP_MAKECALL:
        return make_call(uc, (uc_makecall_t *) data);
    case UC_OP_SENDMOREINFO:
        return send_more_info(uc, call, (uc_sendmoreinfo_t *) data);
    case UC_OP_CALLACK:
        return call_ack(uc, call, (uc_callack_t *) data);
    case UC_OP_ACCEPTCALL:
        return accept_call(uc, call);
    case UC_OP_ANSWERCALL:
        return answer_call(uc, call);
    case UC_OP_DROPCALL:
        return drop_call(uc, call, (int) (intptr_t) data);
    case UC_OP_RELEASECALL:
        return release_call(uc, call);
    case UC_OP_UNBLOCK:
        return unblock(uc, (int) (intptr_t) data);
    case UC_OP_BLOCK:
        return block(uc, (int) (intptr_t) data);
    case UC_OP_USERTOUSER:
        return user_to_user(uc, call, (uc_usertouser_t *) data);
    }
    /*endswitch*/
    return  UC_RET_UNSUPPORTED;
}
/*- End of function --------------------------------------------------------*/

uc_t *create_new(int fd, const char *variant, int mode, int outgoing_calls_allowed)
{
    uc_t *uc;
    fx_signaling_state_t *fx;
    uc_chan_t *chans;
    struct zt_params p;
    struct zt_bufferinfo b;
    int variation;
    int sigtype;
    int x;
    int ch;
    int len;
    const char *s;
    char *t;

    uc_log(NULL, UC_LOG_FLOW, "create_new\n");
    if ((variation = protocol_variant(variant)) < 0)
    {
        uc_error("Protocol variant '%s' unknown\n", variant);
        return  NULL;
    }
    /*endif*/
    if (ioctl(fd, ZT_GET_PARAMS, &p))
    {
        uc_error("Unable to get channel parameters: %s\n", strerror(errno));
        return  NULL;
    }
    /*endif*/
    switch (variation)
    {
    case FX_PROT_LS:
        sigtype = (mode == UC_MODE_CPE)  ?  ZT_SIG_FXSLS  :  ZT_SIG_FXOLS;
        break;
    case FX_PROT_GS:
        sigtype = (mode == UC_MODE_CPE)  ?  ZT_SIG_FXSGS  :  ZT_SIG_FXOGS;
        break;
    case FX_PROT_KS:
        sigtype = (mode == UC_MODE_CPE)  ?  ZT_SIG_FXSKS  :  ZT_SIG_FXOKS;
        break;
    }
    /*endswitch*/
    if (p.sigtype != sigtype)
    {
        uc_error("In signaling mode %d, instead of %d.\n", p.sigtype, sigtype);
        return  NULL;
    }
    /*endif*/
    if ((uc = malloc(sizeof(*uc))) == NULL)
    {
        uc_error("Failed to allocate buffers.\n");
        return  NULL;
    }
    /*endif*/
    memset(uc, '\0', sizeof(*uc));
    uc->protocol_variant = variation;
    uc->protocol_mode = mode;

    if ((fx = malloc(sizeof(*fx))) == NULL)
    {
        uc_error("Failed to allocate buffers.\n");
        free(uc);
        return  NULL;
    }
    /*endif*/
    memset(fx, '\0', sizeof(*fx));
    uc->signaling = fx;
    fx->outgoing_calls_allowed = outgoing_calls_allowed;
    fx->dropped = FALSE;
    
    fx->seize_wait_timer =
    fx->local_unblocking_timer =
    fx->wait_for_answer_timer = -1;

    fx->seize_wait = 200;
    
    s = variant;
    len = variant_element(&s, 1);
    if (len == 0)
    {
        fx->super_tones = get_supervisory_tone_set("hk");
    }
    else
    {
        t = malloc(len + 1);
        strncpy(t, s, len);
        t[len] = '\0';
        fx->super_tones = get_supervisory_tone_set(t);
        free(t);
    }
    /*endif*/
    if (fx->super_tones == NULL)
    {
        uc_error("Failed to load supervisory tones.\n");
        free(fx);
        free(uc);
        return  NULL;
    }
    /*endif*/
    s = variant;
    len = variant_element(&s, 2);
    if (len  &&  s[0] == 'p')
        fx->pulse_dial = TRUE;
    /*endif*/
    s = variant;
    len = variant_element(&s, 3);
    if (len)
    {
        if (strcmp(s, "class") == 0)
            fx->adsi_tx_state.standard = ADSI_STANDARD_CLASS;
        else if (strcmp(s, "clip") == 0)
            fx->adsi_tx_state.standard = ADSI_STANDARD_CLIP;
        else if (strcmp(s, "aclip") == 0)
            fx->adsi_tx_state.standard = ADSI_STANDARD_ACLIP;
        else if (strcmp(s, "jclip") == 0)
            fx->adsi_tx_state.standard = ADSI_STANDARD_JCLIP;
        else if (strcmp(s, "clip-dtmf") == 0)
            fx->adsi_tx_state.standard = ADSI_STANDARD_CLIP_DTMF;
        else
            fx->adsi_tx_state.standard = ADSI_STANDARD_NONE;
        uc_log(NULL, UC_LOG_FLOW, "Caller ID standard %d\n", fx->adsi_tx_state.standard);
    }
    /*endif*/

    if ((chans = malloc(sizeof(*chans))) == NULL)
    {
        uc_error("Failed to allocate buffers.\n");
        free(fx);
        free(uc);
        return  NULL;
    }
    /*endif*/
    memset(chans, '\0', sizeof(*chans));
    uc->num_channels = 1;
    uc->chan = chans;
    ch = 0;
    uc->chan[ch].active_rx =
    uc->chan[ch].active_tx = ACTIVE_NONE;
    uc->chan[ch].handle = fd;
    uc->chan[ch].state = UC_STATE_BLOCKED;
    uc->chan[ch].logging_level = 0;
    uc->chan[ch].native_law =
    uc->chan[ch].current_law = p.curlaw;
    uc->chan[ch].native_codec =
    uc->chan[ch].api_codec = (p.curlaw == ZT_LAW_ALAW)  ?  UC_CODEC_ALAW  :  (p.curlaw == ZT_LAW_MULAW)  ?  UC_CODEC_ULAW  :  UC_CODEC_UNKNOWN;

    if (ioctl(fd, ZT_GET_BUFINFO, &b))
    {
        free(chans);
        free(fx);
        free(uc);
        uc_error("Unable to get channel buffer info: %s\n", strerror(errno));
        return  NULL;
    }
    /*endif*/
    uc->chan[ch].audio_bufsize = b.bufsize;

    hook_control(uc, ch, TRUE);
    hash_InitHashTable(&uc->calls, HASH_ONE_WORD_KEYS);
    return uc;
}
/*- End of function --------------------------------------------------------*/

static int channel_open(const char *variant, char *dev)
{
    struct zt_bufferinfo bi;
    struct zt_gains g;
    int x;
    int fd;
    int i;
    int chan;

    if ((chan = atoi(dev)) > 0)
    {
        if ((fd = open("/dev/zap/channel", O_RDWR | O_NONBLOCK)) < 0)
            return -1;
        /*endif*/
        if (ioctl(fd, ZT_SPECIFY, &chan))
        {
            x = errno;
            close(fd);
            errno = x;
            return -1;
        }
        /*endif*/
    }
    else
    {
        if ((fd = open(dev, O_RDWR | O_NONBLOCK)) < 0)
            return -1;
        /*endif*/
    }
    /*endif*/
    if (ioctl(fd, ZT_GET_BUFINFO, &bi) < 0)
    {
        x = errno;
        close(fd);
        errno = x;
        return -1;
    }
    /*endif*/
    bi.txbufpolicy = ZT_POLICY_IMMEDIATE;
    bi.rxbufpolicy = ZT_POLICY_IMMEDIATE;
    bi.numbufs = 4;
    bi.bufsize = 160;
    if (ioctl(fd, ZT_SET_BUFINFO, &bi) < 0)
    {
        x = errno;
        close(fd);
        errno = x;
        return -1;
    }
    /*endif*/
    if (ioctl(fd, ZT_CHANNO, &chan))
    {
        x = errno;
        close(fd);
        errno = x;
        return -1;
    }
    /*endif*/
    /* Set default gains */
    g.chan = 0;
    for (i = 0;  i < 256;  i++)
    {
        g.rxgain[i] = i;
        g.txgain[i] = i;
    }
    if (ioctl(fd, ZT_SETGAINS, &g) < 0)
    {
        x = errno;
        close(fd);
        errno = x;
        return -1;
    }
    /*endif*/
    return fd;
}
/*- End of function --------------------------------------------------------*/

static int channel_close(int fd)
{
    close(fd);
    return 0;
}
/*- End of function --------------------------------------------------------*/

static char *error_message(uc_t *uc, int code)
{
    switch (code)
    {
    case FX_PROTOCOL_FAIL_DEVICE_IO_ERROR:
        return "Device I/O error";
    }
    /*endswitch*/
    return "Undefined";
}
/*- End of function --------------------------------------------------------*/

static int get_device_handles(uc_t *uc, int ch, int *fe, int *be)
{
    if (fe)
        *fe = uc->chan[ch].handle;
    /*endif*/
    if (be)
        *be = uc->chan[ch].handle;
    /*endif*/
    return UC_RET_OK;
}
/*- End of function --------------------------------------------------------*/

static int channel_set_api_codec(uc_t *uc, int ch, int codec)
{
    int linear;
    int law;

    //uc_log(uc, UC_LOG_FLOW, "Channel set API codec - %d\n", codec);
    switch (codec)
    {
    case UC_CODEC_DEFAULT:
        linear = 0;
        if (ioctl(uc->chan[ch].handle, ZT_SETLINEAR, &linear))
        {
            if (uc->chan[ch].channel_error)
                uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, FX_PROTOCOL_FAIL_DEVICE_IO_ERROR);
            /*endif*/
            return UC_RET_DEVICE_ERROR;
        }
        /*endif*/
        law = uc->chan[ch].native_law;
        if (law == ZT_LAW_ALAW)
            codec = UC_CODEC_ALAW;
        else if (law == ZT_LAW_MULAW)
            codec = UC_CODEC_ULAW;
        /*endif*/
        break;
    case UC_CODEC_LINEAR16:
        linear = 1;
        if (ioctl(uc->chan[ch].handle, ZT_SETLINEAR, &linear))
        {
            if (uc->chan[ch].channel_error)
                uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, FX_PROTOCOL_FAIL_DEVICE_IO_ERROR);
            /*endif*/
            return UC_RET_DEVICE_ERROR;
        }
        /*endif*/
        law = uc->chan[ch].native_law;
        break;
    case UC_CODEC_ALAW:
        linear = 0;
        if (ioctl(uc->chan[ch].handle, ZT_SETLINEAR, &linear))
        {
            if (uc->chan[ch].channel_error)
                uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, FX_PROTOCOL_FAIL_DEVICE_IO_ERROR);
            /*endif*/
            return UC_RET_DEVICE_ERROR;
        }
        /*endif*/
        law = ZT_LAW_ALAW;
        break;
    case UC_CODEC_ULAW:
        linear = 0;
        if (ioctl(uc->chan[ch].handle, ZT_SETLINEAR, &linear))
        {
            if (uc->chan[ch].channel_error)
                uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, FX_PROTOCOL_FAIL_DEVICE_IO_ERROR);
            /*endif*/
            return UC_RET_DEVICE_ERROR;
        }
        /*endif*/
        law = ZT_LAW_MULAW;
        break;
    default:
        return  UC_RET_UNSUPPORTED;
    }
    /*endswitch*/
    if (uc->chan[ch].current_law != law)
    {
        uc->chan[ch].current_law = law;
        if (ioctl(uc->chan[ch].handle, ZT_SETLAW, &law))
        {
            if (uc->chan[ch].channel_error)
                uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, FX_PROTOCOL_FAIL_DEVICE_IO_ERROR);
            /*endif*/
            return UC_RET_DEVICE_ERROR;
        }
        /*endif*/
    }
    /*endif*/
    uc->chan[ch].api_codec = codec;
    return  UC_RET_OK;
}
/*- End of function --------------------------------------------------------*/

static int channel_write(uc_t *uc, int ch, uint8_t *buf, int len)
{
    int xlen;

    if (uc->chan[ch].active_tx != ACTIVE_SPEECH)
        return len;
    /*endif*/
    if (uc->chan[ch].channel_write_codec)
        len = uc->chan[ch].channel_write_codec(uc, ch, uc->chan[ch].channel_write_codec_user_data, buf, len);
    /*endif*/
    for (;;)
    {
        xlen = write(uc->chan[ch].handle, buf, len);
        if (xlen >= len)
            break;
        /*endif*/
        if (xlen < 0)
        {
            if (errno == ELAST)
            {
                check_event(uc);
                continue;
            }
            /*endif*/
            if (errno == EAGAIN)
                continue;
            /*endif*/
            if (uc->chan[ch].channel_error)
                uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, FX_PROTOCOL_FAIL_DEVICE_IO_ERROR);
            /*endif*/
            break;
        }
        /*endif*/
        len -= xlen;
        buf += len;
    }
    /*endfor*/
    return  xlen;
}
/*- End of function --------------------------------------------------------*/

static int channel_flush(uc_t *uc, int ch, int which)
{
    fx_signaling_state_t *fx;
    int x;

    if (ch >= uc->num_channels)
        return UC_RET_BAD_PARAMETER;
    /*endif*/
    fx = (fx_signaling_state_t *) uc->signaling;
    uc_log(uc, UC_LOG_FLOW, "Channel flush\n");
    if ((which & 1))
    {
        x = ZT_FLUSH_WRITE;
        if (ioctl(uc->chan[ch].handle, ZT_FLUSH, &x))
        {
            if (uc->chan[ch].channel_error)
                uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, FX_PROTOCOL_FAIL_DEVICE_IO_ERROR);
            /*endif*/
            return  UC_RET_DEVICE_ERROR;
        }
        /*endif*/
    }
    /*endif*/
    if ((which & 2))
    {
        /* TODO: implement input flush */
    }
    /*endif*/
    return  UC_RET_OK;
}
/*- End of function --------------------------------------------------------*/

int channel_gains(uc_t *uc, int ch, float rxgain, float txgain)
{
    fx_signaling_state_t *fx;
    struct zt_gains g;
    float gain;
    int j;
    int k;

    if (ch >= uc->num_channels)
        return UC_RET_BAD_PARAMETER;
    /*endif*/
    fx = (fx_signaling_state_t *) uc->signaling;
    uc_log(uc, UC_LOG_FLOW, "Channel gains\n");
    g.chan = 0;
    if (rxgain != 0.0)
    {
        /* Calculate linear value of rx gain */
        gain = pow(10.0, rxgain/20.0);
        if (uc->chan[ch].current_law == UC_CODEC_ALAW)
        {
            for (j = 0;  j < 256;  j++)
            {
                k = fsaturate(((float) alaw_to_linear(j))*gain);
                g.rxgain[j] = linear_to_alaw(k);
            }
            /*endfor*/
        }
        else
        {
            for (j = 0;  j < 256;  j++)
            {
                k = fsaturate(((float) ulaw_to_linear(j))*gain);
                g.rxgain[j] = linear_to_ulaw(k);
            }
            /*endfor*/
        }
        /*endif*/
    }
    else
    {
        for (j = 0;  j < 256;  j++)
            g.rxgain[j] = j;
        /*endfor*/
    }
    /*endif*/
    if (txgain != 0.0)
    {
        /* Calculate linear value of tx gain */
        gain = pow(10.0, txgain/20.0);
        if (uc->chan[ch].current_law == UC_CODEC_ALAW)
        {
            for (j = 0;  j < 256;  j++)
            {
                k = fsaturate(((float) alaw_to_linear(j))*gain);
                g.txgain[j] = linear_to_alaw(k);
            }
            /*endfor*/
        }
        else
        {
            for (j = 0;  j < 256;  j++)
            {
                k = fsaturate(((float) ulaw_to_linear(j))*gain);
                g.txgain[j] = linear_to_ulaw(k);
            }
            /*endfor*/
        }
        /*endif*/
    }
    else
    {
        for (j = 0;  j < 256;  j++)
            g.txgain[j] = j;
        /*endfor*/
    }
    /*endif*/
    if (ioctl(uc->chan[ch].handle, ZT_SETGAINS, &g))
    {
        if (uc->chan[ch].channel_error)
            uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, FX_PROTOCOL_FAIL_DEVICE_IO_ERROR);
        /*endif*/
        return  UC_RET_DEVICE_ERROR;
    }
    /*endif*/
    return  UC_RET_OK;
}
/*- End of function --------------------------------------------------------*/

int channel_echo_cancel(uc_t *uc, int ch, int op)
{
    fx_signaling_state_t *fx;
    int x;

    if (ch >= uc->num_channels)
        return UC_RET_BAD_PARAMETER;
    /*endif*/
    fx = (fx_signaling_state_t *) uc->signaling;
    uc_log(uc, UC_LOG_FLOW, "Channel echo cancel\n");
    switch (op)
    {
    case UC_ECHO_CANCEL_OFF:
    case UC_ECHO_CANCEL_ON:
        x = op;
        if (ioctl(uc->chan[ch].handle, ZT_ECHOCANCEL, &x) < 0)
        {
            if (uc->chan[ch].channel_error)
                uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, FX_PROTOCOL_FAIL_DEVICE_IO_ERROR);
            /*endif*/
            return  UC_RET_DEVICE_ERROR;
        }
        /*endif*/
        break;
    case UC_ECHO_CANCEL_NOTRAINING:
    case UC_ECHO_CANCEL_TRAINING:
        x = op - UC_ECHO_CANCEL_NOTRAINING;
        if (ioctl(uc->chan[ch].handle, ZT_ECHOTRAIN, &x) < 0)
        {
            if (uc->chan[ch].channel_error)
                uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, FX_PROTOCOL_FAIL_DEVICE_IO_ERROR);
            /*endif*/
            return  UC_RET_DEVICE_ERROR;
        }
        /*endif*/
        break;
    default:
        return UC_RET_BAD_PARAMETER;
    }
    /*endswitch*/
    return  UC_RET_OK;
}
/*- End of function --------------------------------------------------------*/

int channel_switching(uc_t *uc, int ch, int op, int dest, int parms)
{
    struct zt_confinfo zi;
    fx_signaling_state_t *fx;
    int x;

    if (ch >= uc->num_channels)
        return UC_RET_BAD_PARAMETER;
    /*endif*/
    fx = (fx_signaling_state_t *) uc->signaling;
    uc_log(uc, UC_LOG_FLOW, "Channel switching\n");
    switch (op)
    {
    case UC_SWITCHING_FREE:
        memset(&zi, 0, sizeof(zi));
        if (ioctl(uc->chan[ch].handle, ZT_SETCONF, &zi))
        {
            if (uc->chan[ch].channel_error)
                uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, FX_PROTOCOL_FAIL_DEVICE_IO_ERROR);
            /*endif*/
            return  UC_RET_DEVICE_ERROR;
        }
        /*endif*/
        break;
    case UC_SWITCHING_UNMUTE:
    case UC_SWITCHING_MUTE:
        x = op - UC_SWITCHING_UNMUTE;
        if (ioctl(uc->chan[ch].handle, ZT_CONFMUTE, &x))
        {
            if (uc->chan[ch].channel_error)
                uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, FX_PROTOCOL_FAIL_DEVICE_IO_ERROR);
            /*endif*/
            return  UC_RET_DEVICE_ERROR;
        }
        /*endif*/
        break;
    case UC_SWITCHING_CONNECT:
        memset(&zi, 0, sizeof(zi));
        zi.confno = dest;
        zi.confmode = parms;
        if (ioctl(uc->chan[ch].handle, ZT_SETCONF, &zi))
        {
            if (uc->chan[ch].channel_error)
                uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, FX_PROTOCOL_FAIL_DEVICE_IO_ERROR);
            /*endif*/
            return  UC_RET_DEVICE_ERROR;
        }
        /*endif*/
        break;
    default:
        return UC_RET_BAD_PARAMETER;
    }
    /*endswitch*/
    return  UC_RET_OK;
}
/*- End of function --------------------------------------------------------*/

static int hook_control(uc_t *uc, int ch, int on_hook)
{
    int x;

    x = (on_hook)  ?  ZT_ONHOOK  :  ZT_OFFHOOK;
    if (ioctl(uc->chan[ch].handle, ZT_HOOK, &x))
    {
        if (errno != EINPROGRESS)
        {
           if (uc->chan[ch].channel_error)
                uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, FX_PROTOCOL_FAIL_DEVICE_IO_ERROR);
            /*endif*/
            return  UC_RET_DEVICE_ERROR;
        }
        /*endif*/
    }
    /*endif*/
    return  UC_RET_OK;
}
/*- End of function --------------------------------------------------------*/

static int pulse_dial(uc_t *uc, int ch, const char *digits, int replace)
{
    ZT_DIAL_OPERATION dop;

    /* TODO: check the right value for not replacing */
    dop.op = (replace)  ?  ZT_DIAL_OP_REPLACE  :  ZT_DIAL_OP_APPEND;
    if (digits[0] == '\0')
        dop.op = ZT_DIAL_OP_CANCEL;
    snprintf(dop.dialstr, sizeof(dop.dialstr), "P%s", digits);
    if (ioctl(uc->chan[ch].handle, ZT_DIAL, &dop))
    {
        if (errno != EINPROGRESS)
        {
            if (uc->chan[ch].channel_error)
                uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, FX_PROTOCOL_FAIL_DEVICE_IO_ERROR);
            /*endif*/
            return  UC_RET_DEVICE_ERROR;
        }
        /*endif*/
    }
    /*endif*/
    return  UC_RET_OK;
}
/*- End of function --------------------------------------------------------*/

static int ring_control(uc_t *uc, int ch, int ring_on)
{
    int x;

    x = (ring_on)  ?  ZT_RING  :  ZT_RINGOFF;
    if (ioctl(uc->chan[ch].handle, ZT_HOOK, &x))
    {
        if (errno != EINPROGRESS)
        {
            if (uc->chan[ch].channel_error)
                uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, FX_PROTOCOL_FAIL_DEVICE_IO_ERROR);
            /*endif*/
            return  UC_RET_DEVICE_ERROR;
        }
        /*endif*/
    }
    /*endif*/
    return  UC_RET_OK;
}
/*- End of function --------------------------------------------------------*/

static int control_ring_cadence(uc_t *uc, int ch, int *cadence)
{
    if (ioctl(uc->chan[ch].handle, ZT_SETCADENCE, cadence))
    {
        if (errno != EINPROGRESS)
        {
            uc_log(uc, UC_LOG_WARNING, "Unable to set ring cadence.\n");
            if (uc->chan[ch].channel_error)
                uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, FX_PROTOCOL_FAIL_DEVICE_IO_ERROR);
            /*endif*/
            return  UC_RET_DEVICE_ERROR;
        }
        /*endif*/
    }
    /*endif*/
    return UC_RET_OK;
}
/*- End of function --------------------------------------------------------*/

static int control_echo_cancelling(uc_t *uc, int ch, int mode)
{
    int x;

    x = 0;
    if (ioctl(uc->chan[ch].handle, ZT_ECHOCANCEL, &x))
    {
        if (errno != EINPROGRESS)
        {
            uc_error("Unable to set echo canceller: %s\n", strerror(errno));
            if (uc->chan[ch].channel_error)
                uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, FX_PROTOCOL_FAIL_DEVICE_IO_ERROR);
            /*endif*/
            return  UC_RET_DEVICE_ERROR;
        }
        /*endif*/
    }
    /*endif*/
    return UC_RET_OK;
}
/*- End of function --------------------------------------------------------*/

static int check_alarms(uc_t *uc, int ch, int *alarms)
{
    ZT_SPANINFO zi;

    memset(&zi, 0, sizeof(zi));
    zi.spanno = 0;
    if (ioctl(uc->chan[ch].handle, ZT_SPANSTAT, &zi) < 0)
    {
        uc_error("Unable to get alarm status: %s\n", strerror(errno));
        if (uc->chan[ch].channel_error)
            uc->chan[ch].channel_error(uc, ch, uc->chan[ch].channel_error_user_data, FX_PROTOCOL_FAIL_DEVICE_IO_ERROR);
        /*endif*/
        return  UC_RET_DEVICE_ERROR;
    }
    /*endif*/
    *alarms = zi.alarms;
    return UC_RET_OK;
}
/*- End of function --------------------------------------------------------*/
/*- End of file ------------------------------------------------------------*/
